/**
 * BigFraction - for arbitrary precision fractional numbers
 * Version 1.00 - March 27th 2007
 * 
 * * Copyright (c) 2007, Steven Goodwin.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the <organization> nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Steven Goodwin ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.bluedust.math;
import java.math.BigInteger;

/**
 * The class represents an immutable fraction, as held in numerator 
 * and denominator. The fraction may be top-heavy, but is reduced when
 * printed.
 * 
 * The value can be converted into float's if desired, although precision
 * will invariably be lost, and some large fractions (such as calculations for
 * pi) can not be converted if the values of BigInteger exceed the range capable of
 * floats.
 * 
 * @author steev
 *
 */

public class BigFraction {
	BigInteger numerator;
	BigInteger denominator;
	
	public BigFraction(int n) {
		numerator = BigInteger.valueOf(n);
		denominator = BigInteger.ONE;
	}
	
	public BigFraction(int n, int d) {
		numerator = BigInteger.valueOf(n);
		denominator = BigInteger.valueOf(d);
	}
	
	public BigFraction(BigInteger n, BigInteger d) {
		numerator = new BigInteger(n.toByteArray());
		denominator = new BigInteger(d.toByteArray());
	}
	
	public BigFraction(BigFraction f) {
		this(f.numerator, f.denominator);
	}
	
	public BigInteger getNominator() {
		reduceThis();
		return numerator;
	}
	
	public BigInteger getDenominator() {
		reduceThis();
		return denominator;
	}
	
	public BigFraction add(BigFraction rhs) {
		// a   b      a.d + b.c
		// - + -  =   ---------
		// c   d         c.d
		// a/c = this
		// b/d = rhs
		BigInteger ad = numerator.multiply(rhs.denominator);
		BigInteger bc = denominator.multiply(rhs.numerator);
		BigInteger cd = denominator.multiply(rhs.denominator);
		return new BigFraction(ad.add(bc), cd);
	}
	
	public BigFraction subtract(BigFraction rhs) {
		BigInteger ad = numerator.multiply(rhs.denominator);
		BigInteger bc = denominator.multiply(rhs.numerator);
		BigInteger cd = denominator.multiply(rhs.denominator);
		return new BigFraction(ad.subtract(bc), cd);
	}
	
	public BigFraction multiply(BigFraction rhs) {
		return new BigFraction(numerator.multiply(rhs.numerator), denominator.multiply(rhs.denominator));
	}
	
	public BigFraction divide(BigFraction rhs) {
		// we turn the fraction upside down and multiply
		return new BigFraction(numerator.multiply(rhs.denominator), denominator.multiply(rhs.numerator));
	}
	
	public BigFraction min(BigFraction rhs) {
		return compareTo(rhs) < 0 ? new BigFraction(this) : new BigFraction(rhs);
	}
	
	public BigFraction max(BigFraction rhs) {
		return compareTo(rhs) > 0 ? new BigFraction(this) : new BigFraction(rhs);
	}
	
	public BigFraction abs() {
		return new BigFraction(numerator.abs(), denominator.abs());
	}
	
	public BigFraction reciprocal() {
		return new BigFraction(denominator, numerator);
	}
	
	public BigFraction negate() {
		reduceThis();
		return new BigFraction(numerator.negate(), denominator);
	}
	
	public int compareTo(BigFraction rhs) {
		// There's no LCM in BigInteger, so we get identical denominators
		// by multipling them together.
		
		// We're not reducing since that appeared to be more
		// costly than just doing the multiplication on (potentially) larger
		// numbers.
		
		BigInteger lhsNominator = numerator.multiply(rhs.denominator);
		BigInteger rhsNominator = denominator.multiply(rhs.numerator);
		// the denomintor for both fractions is denom*rhs.denom, but we
		// don't need to compute that
		
		return lhsNominator.compareTo(rhsNominator);
	}
	
	public boolean equals(BigFraction rhs) {
		return compareTo(rhs) == 0 ? true : false;
	}
	
	public static String toString(BigFraction f) {
		return f.toString();
	}
	
	public String toString() {
		reduceThis();
		String str = "";
		
		// Is it negative? The reduction ensures only the nom is -ve, and is handled
		// by either:
		// 1. the whole number
		// 2. the nom in the purely fractional version
		// Therefore we need to abs dar[1] later
		
		// Is the fraction top-heavy? If so, we need to introduce a whole number
		BigInteger[]  dar = numerator.divideAndRemainder(denominator);
		
		if (!dar[0].equals(BigInteger.ZERO)) {
			str += dar[0].toString()+" ";
			// Calculate the new nominator; which not coincidentally
			// is the remainder
			str += dar[1].abs().toString();
		} else {
			str += numerator.toString();
		}

		return str + "/" + denominator.toString();
	}
	
	public int intValue() {
		reduceThis();
		return numerator.divide(denominator).intValue();
	}
	
	public long longValue() {
		reduceThis();
		return numerator.divide(denominator).longValue();
	}
	
	public float floatValue() {
		// TODO When we get BigFraction we can convert numerator
		// and denominator into that, before calculating the
		// division, and returning the result.
		reduceThis();
		return numerator.floatValue() / denominator.floatValue();
	}
	
	public double doubleValue() {
		// TODO See note in @floatValue
		reduceThis();
		return numerator.doubleValue() / denominator.doubleValue();
	}
	
	public BigFraction reduce() {
		BigInteger newNumerator, newDenominator;
		
		if (numerator.signum() == denominator.signum()) {
			newNumerator = numerator.abs();
			newDenominator = denominator.abs();
		} else {
			newNumerator = numerator.abs().negate();
			newDenominator = denominator.abs();
		}
		BigInteger gdc = newNumerator.gcd(newDenominator);
		return new BigFraction(newNumerator.divide(gdc), newDenominator.divide(gdc));
	}
	
	public void reduceThis() {
		BigFraction reduced = reduce();

		numerator = reduced.numerator;
		denominator = reduced.denominator;
	}
}
